import numpy as np
from abaqus_parse.mesh import (
    make_donut_mesh,
    make_fine_plus_donut_mesh,
    compact_tension_specimen_mesh
)


def generate_compact_tension_specimen_parts(dimension, mesh_definition,
                     elem_type, size_type, fraction, specimen_material):

    """
    Parameters
    -----------
    dimension: string
        Specification of the number of dimensions: '2D' or '3D'.
    """

    refined_mesh_definition = make_fine_plus_donut_mesh(
        mesh_definition['crack_tip_radius_microns'],
        mesh_definition['fine_mesh_length'],
        mesh_definition['fine_mesh_element_length'],
        mesh_definition['fan_box_num_side_elements'],
        mesh_definition['fan_box_width'],
        ret_crack_definition=True,
        size_behind_crack=0.2
    )

    return {
        'ct-specimen': {
            **compact_tension_specimen_mesh(refined_mesh_definition, dimension, size_type, fraction),
            'element_type': elem_type,
            'sections':[
                {
                    'type': 'Solid',
                    'material': specimen_material,
                    'elset': 'specimen'    
                },
                {
                    'type': 'Solid',
                    'material': 'rigid',
                    'elset': 'ridge'
                }
            ],
        }
    }
    


import numpy as np
from more_itertools import pairwise
from abaqus_parse import fracture, utils

def donut_polar_grid_quad(inner_r, outer_r, num_rays, num_arcs, quad=1,     include_borders=True):
    """
    Create a donut polar grid of nodes for one quadrant.
    
    Parameters
    ----------
    inner_r : float
        Size of inner radius of donut
    outer_r : float
        Size of outer radius of donut
    num_rays : int
        Number of rays within a quadrant of the circle (splits angles)
    num_arcs : int
        Number of arcs within a quadrant of the circle (splits radius)
    quad : int
        Circle quadrant to be populated, accepted values are 1, 2, 3 or 4 (default is 1)
    include_borders : bool
        Include border nodes. Default value is True.    
    
    Outputs
    -------
    tuple (ndarray, ndarray), shape=(N,)
        Polar coordinates distance-angle for nodes.
        
    
    """
    if quad == 1:
        ang_lim_low = 0.0
        ang_lim_upp = np.pi / 2 + 0.1
        
    r = np.linspace(inner_r, outer_r, num_arcs )
    theta = np.linspace(0, 2 * np.pi, (num_rays - 1) * 4 +1)

    radius_matrix, theta_matrix = np.meshgrid(r,theta)
    quat_idx = np.intersect1d(np.where(theta_matrix.flatten() >= ang_lim_low)[0],
               np.where(theta_matrix.flatten() <= ang_lim_upp)[0])
    
    
    return theta_matrix.flatten()[quat_idx], radius_matrix.flatten()[quat_idx]


def order_nodes_elem(elem_nodes_all, all_nodes, all_nodes_labs):
    """
    Order the nodes of an element in anticlockwise direction, consistent with the element definition in Abaqus.
    
    Parameters:
    elem_nodes_all : array of (M, n)
        The nodes that form part of M elements with n corners each.
    all_nodes : array of (N, 2)
        The coordinates of all N nodes in the mesh.
    all_nodes_labs : array of (N,)
        The labels of all nodes in the mesh.
    
    """
    elem_nodes_all_ordered = []
    for el_idx, elem_nodes in enumerate(elem_nodes_all):
        
        
        # # sort nodes in growing order
        # node_el_sorted = np.sort(elem_nodes)
        
        # indices of nodes in node list
        node_idx = np.where(np.isin(all_nodes_labs, elem_nodes))[0]
       
        if len(node_idx)==3:
            vals, counts = np.unique(elem_nodes, return_counts=True)
            rep_idx = np.where(counts == 2)[0]
            node_idx = np.insert(node_idx, rep_idx, node_idx[np.where(counts == 2)])

            
        node_el_sorted_by_node_list = all_nodes_labs[node_idx]    
        # coords of nodes in the order they are in the all nodes list?
        node_coord_sorted = all_nodes[node_idx]
        # order nodes in correct order
        
        node_ord = utils.order_coplanar_points(
            np.hstack((node_coord_sorted, np.zeros((len(node_coord_sorted),1)))).T,
            np.array([[0,0,1]]).T
        )
        
        elem_nodes_all_ordered.append(node_el_sorted_by_node_list[node_ord])
        
    return np.concatenate(elem_nodes_all_ordered).reshape(-1, 4)



def cells_from_nodes(nodes_array):
    """
    Find the cell vertices in a regular grid defined by an array of nodes
    
    """
    c = 0
    cells = []

    for ri, row in enumerate(nodes_array):
        count = 0
        if ri != len(nodes_array)-1:
            cells.append([])
        for item1, item2 in pairwise(row):
            if ri == 0:
                cells[ri].append([item1, item2])
            elif ri == len(nodes_array)-1:
                cells[ri-1][count].append(item1)
                cells[ri-1][count].append(item2)
    
            else:
                cells[ri].append([item1, item2])
                cells[ri-1][count].append(item1)
                cells[ri-1][count].append(item2)
            count += 1
    
    return np.array(cells)


def make_donut_mesh(crack_tip_radius_microns, fine_mesh_element_length=0.025, fan_box_num_side_elements=3, fan_box_width=5):
    """
    
    Parameters
    ----------
    crack_tip_radius_microns : float
        Inner radius in micrometers
    
    fan_box_num_side_elements : integer
     number of arcs (TO CALCULATE LATER) = 3 
    fan_box_width : integer
       aproximate width of fan mesh in multiples of r 
    ret_rin_nodes : bool
        Return the inner nodes compromising the inner radius in anti-clockwise direction. 
       
    Returns
    -------
    
    
    """    
    rad_mm = crack_tip_radius_microns * 1e-3
    
    # width of fan mesh in multiples of elem side
    fbox_side_mult_elem = int((fan_box_width*rad_mm) / fine_mesh_element_length)
    fbox_side_mult_elem = round(fbox_side_mult_elem/3)*3
    
    # side size of fan mesh in mm
    fbox_side_mm = fbox_side_mult_elem * fine_mesh_element_length
    
    # number of rays
    rn = 2 * fbox_side_mult_elem + 1

    thetaout, rout  = donut_polar_grid_quad(rad_mm, fbox_side_mm,
                                            rn, fan_box_num_side_elements)
    
    xout, yout = utils.polar2cart_2D(rout, thetaout)
    
    # Move nodes at ray ends to the sides of a square
    ##################################################
    # coords from 0 to fan_box_num_side_elements of side square box in multiples of ctip radii
    # half coordinates of the end of fan rays crossing square
    end_ray_half_coords =  np.linspace(0, fbox_side_mm, int((rn-1)/2+1))
    # print('end_ray_half_coords: ', end_ray_half_coords)

    end_ray_repeat_coords = (fbox_side_mm * np.ones(rn))[None]
    end_ray_coords = np.concatenate((
        np.concatenate((end_ray_half_coords[:-1],
                    np.flip(end_ray_half_coords)))[None],
            end_ray_repeat_coords
        ), axis=0)

    # The second half of array indices: 
    # 1st half + mid element + next element - 1 for zero based
    # int((rn-1)/2) + 1 + 1 - 1 
    flip_idx = int((rn-1)/2) + 1 - 1
    end_ray_coords[:, :flip_idx] = np.flip(end_ray_coords[:, :flip_idx], axis=0)

    # Array of nodes as (Rows, Cols, (x, y)_node)
    don_reshaped = np.concatenate((xout[None].T, yout[None].T), axis=1).reshape(rn, fan_box_num_side_elements, 2)
    don_reshaped[:,-1,:] = end_ray_coords.T
    don_nodes_flattened = don_reshaped.reshape(rn*fan_box_num_side_elements, 2)
    don_node_labs = np.arange(1, fan_box_num_side_elements * rn + 1).reshape(rn, fan_box_num_side_elements)
    don_node_labs_flattened = don_node_labs.flatten()
    
    don_cells = cells_from_nodes(don_node_labs)
    don_cells_flattened = don_cells.reshape(-1, 4)
    

    don_cells_flattened = order_nodes_elem(don_cells_flattened, don_nodes_flattened,don_node_labs_flattened)

    don_cell_centres = np.mean(
        cells_from_nodes(don_reshaped), axis=2
    )
    don_cell_centres_flattened = don_cell_centres.reshape((fan_box_num_side_elements-1) * (rn-1), 2)

    don_cell_labs = np.arange(1, (fan_box_num_side_elements-1) * (rn-1) + 1).reshape(fan_box_num_side_elements-1, rn-1)
    don_cell_labs_flattened = don_cell_labs.flatten()

    out = {
        'nodes_flattened': don_nodes_flattened,
        'node_labels_flattened': don_node_labs_flattened,
        'cells_flattened': don_cells_flattened,
        'cell_centres_flattened': don_cell_centres_flattened,
        'cell_labels_flattened': don_cell_labs_flattened,
    }

    return out

def make_fine_plus_donut_mesh(crack_tip_radius_microns, fine_mesh_length=0.2, fine_mesh_element_length = 0.025, fan_box_num_side_elements=4,  fan_box_width=5, ret_crack_definition=True, size_behind_crack=0.25):
    """
    Build refined mesh = 4 rectilinear + 1 donut mesh at crack tip
    ''''''''''''''''''''''''''''''''''''''
    '               '    '               '
    '        5      '  4 '        3      '
    '               '    '               '
    '               '    '               '
    '               '    '               '
    '               '    '               '
    ''''''''''''''''''''''''''''''''''''''
    '        6      ' d-1'        2      '
    ''''''''''''''''' '  '               '
                      ''''''''''''''''''''
    
    Parameters
    ----------
    size_behind_crack : float
        Size of the mesh behind the crack as fraction of fine_mesh_length.
    
    """
    rad_mm = crack_tip_radius_microns * 1e-3
    # number of elements on side of fine mesh
    num_elem_side = int(fine_mesh_length / fine_mesh_element_length)
    line_elem =  num_elem_side - 1 #  substract corner element
    line_elem = round(line_elem / 3) * 3
    num_elem_side = line_elem + 1
    ref_mesh_size = [2 * num_elem_side * fine_mesh_element_length, num_elem_side * fine_mesh_element_length]
    
    #refined mesh nodes and cells
    ref_mesh_nodes_all = []
    ref_mesh_labs_all = []
    ref_mesh_cells_all = []
    ref_mesh_cell_centres_all = []
    ref_mesh_cell_labs_all = []
    
    # donut mesh nodes and cells    
    donut_mesh = make_donut_mesh(
        crack_tip_radius_microns,
        fine_mesh_element_length=fine_mesh_element_length, fan_box_num_side_elements=fan_box_num_side_elements, 
        fan_box_width=fan_box_width,
    )
        
    rn = int(donut_mesh['nodes_flattened'].shape[0] / fan_box_num_side_elements)
    don_mesh_size = [
            (donut_mesh['nodes_flattened'].reshape(rn, fan_box_num_side_elements, 2))[0][-1][0],
            (donut_mesh['nodes_flattened'].reshape(rn, fan_box_num_side_elements, 2))[0][-1][0]
    ]
    if ret_crack_definition:
        crack_line = donut_mesh['node_labels_flattened'].reshape(rn, fan_box_num_side_elements)[:,0]
        crack_front = donut_mesh['node_labels_flattened'].reshape(rn, fan_box_num_side_elements)[0]
    
    ref_mesh_nodes_all.append(donut_mesh['nodes_flattened'])
    ref_mesh_labs_all.append(donut_mesh['node_labels_flattened'])
    ref_mesh_cells_all.append(donut_mesh['cells_flattened'])
    ref_mesh_cell_centres_all.append(donut_mesh['cell_centres_flattened'])
    ref_mesh_cell_labs_all.append(donut_mesh['cell_labels_flattened'])


    rect_mesh_shifts = [
        [don_mesh_size[0], 0],
        [don_mesh_size[0], don_mesh_size[1]],
        [0, don_mesh_size[1]],
        [-ref_mesh_size[0] * size_behind_crack, don_mesh_size[1]],
        [-ref_mesh_size[0] * size_behind_crack, 0],
    ]
    rect_mesh_sizes = [
        [ref_mesh_size[0] / 2 - don_mesh_size[0], don_mesh_size[1]],
        [ref_mesh_size[0] / 2 - don_mesh_size[0], ref_mesh_size[1] - don_mesh_size[1]],
        [don_mesh_size[0], ref_mesh_size[1] - don_mesh_size[1]],
        [ref_mesh_size[0] * size_behind_crack, ref_mesh_size[1] - don_mesh_size[1]],
        [ref_mesh_size[0] * size_behind_crack, don_mesh_size[1] - rad_mm]
    ]

    for i in range(5):
        mesh_size = rect_mesh_sizes[i]
        mesh_grid = [int(mesh_size[0]/fine_mesh_element_length) + 1,
                     int(mesh_size[1]/fine_mesh_element_length) + 1]

        w = np.linspace(0, mesh_size[0], mesh_grid[0])  
        h = np.linspace(0, mesh_size[1], mesh_grid[1])

        if i == 4:
            h = list(donut_mesh['nodes_flattened'].reshape(rn, fan_box_num_side_elements, 2)[-1, :, 1])  # DON'T have this
            mesh_grid[1] = len(h)
        meshw, meshh = np.meshgrid(w, h)

        # shift meshgrid 
        meshw += rect_mesh_shifts[i][0]
        meshh += rect_mesh_shifts[i][1]

        # x, y coordinates of nodes
        nodes_coord = np.concatenate([meshw.flatten()[None],
                                      meshh.flatten()[None]], axis=0).T
        node_labs = np.zeros_like(meshw.flatten())
        reshaped = nodes_coord.reshape(mesh_grid[1], mesh_grid[0], 2)
        
        for ni, n in enumerate(nodes_coord):
            # find which nodes already been created 
            cond = np.isclose(n, np.concatenate(ref_mesh_nodes_all), 0.0001)
            found_idx = np.where(np.all(cond, axis=1))        
            if found_idx[0].size>0:
                node_labs[ni] = np.concatenate(ref_mesh_labs_all)[found_idx]

        # Apply mask over repeated border nodes
        mesh_mask = np.zeros_like(meshw)
        if i == 0:
            mesh_mask[:,0] = 1
        elif i == 1:
            mesh_mask[0,:] = 1
        elif i == 2:
            mesh_mask[0,:] = 1
            mesh_mask[:,-1] = 1
        elif i == 3:
            mesh_mask[:,-1] = 1
        elif i == 4:
            mesh_mask[-1,:] = 1
            mesh_mask[:,-1] = 1

        meshw, meshh = [np.ma.masked_array(i, mask=mesh_mask) for i in np.meshgrid(w, h)]
        # shift meshgrid 
        meshw += rect_mesh_shifts[i][0]
        meshh += rect_mesh_shifts[i][1]
        nodes_coord_masked = np.concatenate([meshw.compressed()[None],
                                      meshh.compressed()[None]], axis=0).T

        new_node_idx = np.where(node_labs==0)
        node_labs[new_node_idx] = np.arange(ref_mesh_labs_all[-1][-1]+1,
                                            ref_mesh_labs_all[-1][-1]+1 + len(meshw.compressed()))
        node_labs_grid = node_labs.reshape(meshw.shape)
        if i==0:
            # print('node_labs_grid[0,1:]: ', node_labs_grid[0,1:])
            crack_front = np.concatenate((crack_front, node_labs_grid[0,1:]))
            # print('front: ', front)
        elif i==4:
            crack_lip = node_labs_grid[0,:-1]
            # print('crack_lip: ', crack_lip)
        ref_mesh_nodes_all.append(nodes_coord_masked)
        ref_mesh_labs_all.append(node_labs[new_node_idx])

        # CHANGE!
        cells = (cells_from_nodes(node_labs_grid)).reshape(-1, 4)
        cells = order_nodes_elem(cells, np.concatenate(ref_mesh_nodes_all),
                                 np.concatenate(ref_mesh_labs_all))
        cell_centres = np.mean(
            cells_from_nodes(reshaped), axis=2
        )
        cell_centres_flattened = cell_centres.reshape((mesh_grid[0]-1) * (mesh_grid[1]-1), 2)
        cell_labs = np.arange(ref_mesh_cell_labs_all[-1][-1]+1,
                                            ref_mesh_cell_labs_all[-1][-1]+1 + (mesh_grid[0]-1) * (mesh_grid[1]-1))

        ref_mesh_cells_all.append(cells.astype('int'))
        ref_mesh_cell_centres_all.append(cell_centres_flattened)
        ref_mesh_cell_labs_all.append(cell_labs)
    crack_nodes = {
        'crack_lip': crack_lip.astype('int'),
        'crack_line': crack_line.astype('int'),
        'crack_front': crack_front.astype('int'),
    }

    out = {
        'node_coordinates': ref_mesh_nodes_all,
        'node_labels': ref_mesh_labs_all,
        'element_nodes': ref_mesh_cells_all,
        'element_centre_coordinates': ref_mesh_cell_centres_all,
        'element_labels': ref_mesh_cell_labs_all,
    }
    if ret_crack_definition:
        out['crack_definition'] = crack_nodes

    return out

def find_border(nodes, node_labels, condition):
    """
    Parameters
    ----------
    nodes: ndarray of (N, 2)
    condition: string
        One of minx, maxx, miny, maxy
    
    """
    if condition=='miny':
        nodes_bord = np.min(nodes[:,1])
        nodes_bord_idx =  np.where(nodes[:,1]==nodes_bord)[0]
    elif condition=='maxy':
        nodes_bord = np.max(nodes[:,1])
        nodes_bord_idx =  np.where(nodes[:,1]==nodes_bord)[0]
    elif condition=='minx':
        nodes_bord = np.min(nodes[:,0])
        nodes_bord_idx =  np.where(nodes[:,0]==nodes_bord)[0]
    elif condition=='maxx':
        nodes_bord = np.max(nodes[:,0])
        nodes_bord_idx =  np.where(nodes[:,0]==nodes_bord)[0]
    
    return nodes[nodes_bord_idx], node_labels[nodes_bord_idx]


def expand_mesh(coords, coord_labs, expand_dir, max_node_label, max_elem_label, exp_modes=['transition'], num_layers=[1], corner_beg=False, corner_end=False, end_dir=None, width=1, width_mult=1.5):
    """
    Parameters
    ----------
    coords : (N, 2)
    coord_labs : (N,)
    
    width : float
        Width of a single layer [length].
    
    """
    
    nodes_all = []
    nodes_all_labels = []
    elements_all = []
    elements_all_labels = []
    corner_node_labs = []
    # keep track of the outer most nodes and labels
    coords_outer = np.copy(coords)
    labs_outer = np.copy(coord_labs)
    width_outer = width
    max_node_label = max_node_label
    
    for mi, mode in enumerate(exp_modes):
        n = len(coords_outer)
        vec_outer = expand_dir * width_outer * num_layers[mi]
        if mode=='transition':
            
            # number of blocks of three-to-one elements = 
            nb = (n - 1) // 3
            # number of new outside nodes = N
            N = nb + 1 
            M = (n - 1) % 3      # number of extra nodes (0, 1 or 2)

            vec_inner = 0.5 * vec_outer

            coords_block = coords_outer[:3*nb+1]
            coord_labs_block = labs_outer[:3*nb+1]

            # Expand main line 
            nodes_outer = coords_block[::3] + vec_outer
            nodes_outer_labels = np.arange(max_node_label+1, 
                                            max_node_label+len(nodes_outer)+1)
            max_node_label += len(nodes_outer)
            nodes_inner_a = coords_block[1::3] + vec_inner
            nodes_inner_a_labels = np.arange(max_node_label+1, max_node_label+len(nodes_inner_a)+1)
            max_node_label += len(nodes_inner_a)
            nodes_inner_b = coords_block[2::3] + vec_inner
            nodes_inner_b_labels = np.arange(max_node_label+1, max_node_label+len(nodes_inner_b)+1)
            max_node_label += len(nodes_inner_b)
            
            
            nodes_all_layer = np.concatenate((
                nodes_outer,
                nodes_inner_a,
                nodes_inner_b,
            ))
            nodes_all_labels_layer = np.concatenate((
                nodes_outer_labels,
                nodes_inner_a_labels,
                nodes_inner_b_labels,
            ))
            assert len(set(nodes_all_labels_layer)) == len(nodes_all_labels_layer)
            
            # find elements
            elem_A = np.array([
                coord_labs_block[:-1:3], coord_labs_block[1::3], 
                nodes_inner_a_labels, nodes_outer_labels[:-1]]).T

            elem_B = np.array([
                coord_labs_block[1::3], coord_labs_block[2::3],
                nodes_inner_a_labels, nodes_inner_b_labels
            ]).T

            elem_C = np.array([
                coord_labs_block[2::3], coord_labs_block[::3][1:],
                nodes_inner_b_labels, nodes_outer_labels[1:]
            ]).T

            elem_D = np.array([
                nodes_outer_labels[:-1], nodes_outer_labels[1:],
                nodes_inner_a_labels, nodes_inner_b_labels
            ]).T
            
            elements_all_layer = np.concatenate([
                elem_A,
                elem_B,
                elem_C,
                elem_D
            ])
   
            if M > 0 or corner_end:
                node_end = coords_outer[-1] + vec_outer
                
                
                node_end_lab = max_node_label + 1
                if corner_end:
                    node_end += end_dir * width_outer
                    corner_node_labs.append(node_end_lab)
                    
                # add node to all nodes in layer
                nodes_all_layer = np.concatenate((nodes_all_layer, node_end[None]))
                
                nodes_all_labels_layer = np.concatenate((nodes_all_labels_layer, [node_end_lab]))
                assert len(set(nodes_all_labels_layer)) == len(nodes_all_labels_layer)
                
                # add node to outer layer
                nodes_outer = np.concatenate((nodes_outer, node_end[None]))
                nodes_outer_labels = np.concatenate((nodes_outer_labels, [node_end_lab]))
                
                if M != 2:        
                    elem_end = np.array([labs_outer[-1], coord_labs_block[-1], 
                                node_end_lab, nodes_outer_labels[-2]])[None]
                    elements_all_layer = np.append(elements_all_layer, elem_end, axis=0)
                
                elif M == 2:
                    node_extra = np.mean((nodes_outer[-2], node_end), axis=0)
                    node_extra_lab = node_end_lab + 1
                    nodes_all_layer = np.concatenate((nodes_all_layer, node_extra[None]))
                    
                    nodes_all_labels_layer = np.concatenate((nodes_all_labels_layer, [node_extra_lab]))
                    assert len(set(nodes_all_labels_layer)) == len(nodes_all_labels_layer)
                    
                    nodes_outer = np.concatenate((nodes_outer, node_extra[None]))
                    nodes_outer_labels = np.concatenate((nodes_outer_labels, [node_extra_lab]))

                    elem_extra = np.array([
                        [labs_outer[-2],  labs_outer[-3],
                            nodes_outer_labels[-3], node_extra_lab],
                        [labs_outer[-2], labs_outer[-1],
                            node_extra_lab, node_end_lab]
                    ])
                    elements_all_layer = np.append(elements_all_layer, elem_extra, axis=0)
                    
            if corner_beg:
                node_beg = coords_outer[0] + vec_outer - end_dir * width_outer
                nodes_all_layer = np.concatenate((nodes_all_layer, node_beg[None]))
                node_beg_label = nodes_all_labels_layer[-1]+1
                
                nodes_all_labels_layer = np.append(nodes_all_labels_layer, node_beg_label)
                assert len(set(nodes_all_labels_layer)) == len(nodes_all_labels_layer)
                
                nodes_outer = np.concatenate((nodes_outer, node_beg[None]))
                nodes_outer_labels = np.concatenate((nodes_outer_labels, [node_beg_label]))
                
                elem_beg = np.array([node_beg_label, coord_labs_block[0], 
                                coord_labs_block[0], nodes_outer_labels[0]])[None]
                elements_all_layer = np.append(elements_all_layer, elem_beg, axis=0)
            

        elif mode=='uniform':
            nodes_outer = coords_outer + vec_outer
            nodes_outer_labels = np.arange(max_node_label+1, 
                                            max_node_label+len(nodes_outer)+1)
            if corner_beg:
                node_beg = coords_outer[0] + vec_outer - end_dir * width_outer * num_layers[mi]
                nodes_outer[0] = node_beg
            if corner_end:
                nodes_outer[-1] = nodes_outer[-1] + end_dir * width_outer * num_layers[mi]
                
            nodes_all_layer = np.copy(nodes_outer)
            
            nodes_all_labels_layer = np.copy(nodes_outer_labels)
            assert len(set(nodes_all_labels_layer)) == len(nodes_all_labels_layer)
            
            layers_inner = np.arange(1, num_layers[mi], 1)
        
            max_node_lab_inner = nodes_outer_labels[-1]
            if layers_inner.size > 0:
                nodes_inner = []
                nodes_inner_labels = []
                for layer in layers_inner:
                    vec_inner = expand_dir * width_outer * layer
                    nodes_inner_i = coords_outer + vec_inner
                    if corner_beg:
                        nodes_inner_i[0] = nodes_inner_i[0] - end_dir * width_outer * layer
                    if corner_end:
                        nodes_inner_i[-1] = nodes_inner_i[-1] + end_dir * width_outer * layer
                    nodes_inner.append(nodes_inner_i)
                    nodes_inner_labels.append(
                        np.arange(
                            max_node_lab_inner+1, 
                            max_node_lab_inner+len(coords_outer)+1
                        )
                    )

                    max_node_lab_inner = np.max(np.array(nodes_inner_labels))

                nodes_all_layer = np.concatenate((nodes_all_layer, np.squeeze(np.concatenate(nodes_inner))))
                
                nodes_all_labels_layer = np.append(nodes_all_labels_layer, nodes_inner_labels)
                assert len(set(nodes_all_labels_layer)) == len(nodes_all_labels_layer)
            
            nodes_all_labels_layer_reshaped = nodes_all_labels_layer.reshape(len(nodes_all_labels_layer)//len(coords_outer), len(coords_outer))
            
            labs_outer_sort = labs_outer[coords_outer[:,np.where(end_dir==1)[0][0]].argsort()]
            elem_nodes_concat = np.concatenate((nodes_all_labels_layer_reshaped, labs_outer_sort[None]))
            elements_all_layer = (cells_from_nodes(elem_nodes_concat)).reshape(-1, 4)
            

        elements_all_labels_layer = np.arange(max_elem_label+1, 
                                            max_elem_label+len(elements_all_layer)+1)

        coords_outer = np.copy(nodes_outer)
        labs_outer = np.copy(nodes_outer_labels)
        labs_outer = labs_outer[coords_outer[:,np.where(end_dir==1)[0][0]].argsort()]
        coords_outer = coords_outer[coords_outer[:,np.where(end_dir==1)[0][0]].argsort()]
            
        nodes_all.append(nodes_all_layer)
        nodes_all_labels.append(nodes_all_labels_layer)
        elements_all.append(elements_all_layer)
        elements_all_labels.append(elements_all_labels_layer)
        
        width_outer = (mi + 1) * width_mult * width
        max_node_label = np.max(np.concatenate(nodes_all_labels))
        max_elem_label = np.max(elements_all_labels_layer)

    
    elements_all = order_nodes_elem(np.concatenate(elements_all).astype('int'),
                                    np.concatenate((coords, np.concatenate(nodes_all))), 
                                 np.concatenate((coord_labs, np.concatenate(nodes_all_labels))))
    return np.concatenate(nodes_all), np.concatenate(nodes_all_labels),\
            elements_all, np.concatenate(elements_all_labels)

def expand_mult_dirs(conds, expand_dirs, end_dirs, mesh_nodes_all, 
                    mesh_node_labs_all, mesh_elem_all, mesh_elem_labs_all,
                    num_layers=[3,3], width=1):
    
    for i in range(len(conds)):

        # find max node and element labels to start
        max_node_label = np.max(mesh_node_labs_all)
        max_elem_label = np.max(mesh_elem_labs_all)


        bord_nodes, bord_labs = find_border(
            mesh_nodes_all, 
            mesh_node_labs_all,
            conds[i]
        )    
        sort_idx_x = np.lexsort((bord_nodes[:,1], bord_nodes[:,0],))
        expand_dir = expand_dirs[i]
        end_dir = end_dirs[i]

        expmesh_nodes, expmesh_labels, elements_all, elements_all_labels = expand_mesh(
            bord_nodes[sort_idx_x], 
            bord_labs[sort_idx_x],
            expand_dir, max_node_label=max_node_label,
            max_elem_label=max_elem_label, 
            end_dir=end_dir, width=width,
            exp_modes=[ 'uniform',],
            num_layers=num_layers, 
        )

        mesh_nodes_all = np.concatenate(
            (mesh_nodes_all,
            expmesh_nodes)
        )
        mesh_node_labs_all = np.concatenate(
            (mesh_node_labs_all,
            expmesh_labels)
        )
        mesh_elem_all = np.concatenate(
            (mesh_elem_all,
            elements_all)
        )
        mesh_elem_labs_all = np.concatenate(
            (mesh_elem_labs_all,
            elements_all_labels)
        )
        
    return mesh_nodes_all, mesh_node_labs_all, mesh_elem_all.astype('int'), mesh_elem_labs_all.astype('int')


def transition_mesh(refined_mesh_definition):
    """
    Expand a refined mesh in required directions through a transition mesh with increasing element size. The following joining mesh feature is used:
     _____
    |_|_|_|
    | |_| |
    |/___\|
    
    Parameters
    ----------
    
    
    
    """

    ref_mesh_nodes_all = refined_mesh_definition['node_coordinates']
    ref_mesh_labs_all = refined_mesh_definition['node_labels']
    ref_mesh_cells_all = refined_mesh_definition['element_nodes']
    ref_mesh_cell_centres_all = refined_mesh_definition['element_centre_coordinates']
    ref_mesh_cell_labs_all = refined_mesh_definition['element_labels']
    crack_nodes = refined_mesh_definition['crack_definition']

    ref_mesh_nodes_all_con = np.concatenate(ref_mesh_nodes_all)

    conds = ['minx', 'maxy', 'maxx',]

    min_x = np.min(ref_mesh_nodes_all_con[:,0])
    max_x = np.max(ref_mesh_nodes_all_con[:,0])
    max_y = np.max(ref_mesh_nodes_all_con[:,1])

    # border idx of expansion directions
    ref_mesh_bord_idx = [
        np.where(ref_mesh_nodes_all_con[:,0]==min_x)[0],
        np.where(ref_mesh_nodes_all_con[:,1]==max_y)[0],
        np.where(ref_mesh_nodes_all_con[:,0]==max_x)[0],
    ]

    # expand directions (normal to border)
    expand_dirs = [
        np.array([-1, 0]),
        np.array([0, 1]),
        np.array([1, 0])
    ]

    # start-end direction (parallel to border)
    end_dirs = [
        np.array([0, 1]),
        np.array([1, 0]),
        np.array([0, 1])
    ]
    # corner end beginning and end modes
    corner_end_modes = [True, True, True]
    corner_beg_modes = [False, True, False]

    # create copies of all mesh components as a start
    mesh_nodes_all = ref_mesh_nodes_all.copy()
    mesh_node_labs_all = ref_mesh_labs_all.copy()
    mesh_elem_all = ref_mesh_cells_all.copy()
    mesh_elem_labs_all = ref_mesh_cell_labs_all.copy()

    # find max node and element labels to start
    max_node_label = int(ref_mesh_labs_all[-1][-1])
    max_elem_label = int(ref_mesh_cell_labs_all[-1][-1])

    for i in range(len(conds)):    
        bord_nodes, bord_labs = find_border(ref_mesh_nodes_all_con, 
                                           np.concatenate(ref_mesh_labs_all), conds[i])    
        sort_idx_x = np.lexsort((bord_nodes[:,1], bord_nodes[:,0],))
        expand_dir = expand_dirs[i]
        end_dir = end_dirs[i]

        expmesh_nodes, expmesh_labels, elements_all, elements_all_labels = expand_mesh(bord_nodes[sort_idx_x], 
                                                                                       bord_labs[sort_idx_x],
                                                    expand_dir, max_node_label=max_node_label,
                                                    max_elem_label=max_elem_label, 
                                                    corner_end=corner_end_modes[i], corner_beg=corner_beg_modes[i], 
                                                    end_dir=end_dir, width=0.1,
                                                    exp_modes=['transition', 'uniform', 'transition', 'uniform', 
                                                               'transition', 'uniform', ],
                                                    num_layers=[1, 2, 1, 1, 1, 1], 
                                                                                      )

        max_node_label = expmesh_labels[-1]
        max_elem_label = elements_all_labels[-1]
        mesh_nodes_all.append(
            expmesh_nodes
        )
        mesh_node_labs_all.append(
            expmesh_labels
        )
        mesh_elem_all.append(elements_all)
        mesh_elem_labs_all.append(elements_all_labels)

    mesh_nodes_all = np.concatenate(mesh_nodes_all)
    mesh_node_labs_all = np.concatenate(mesh_node_labs_all).astype('int')
    mesh_elem_all = np.concatenate(mesh_elem_all)
    mesh_elem_labs_all = np.concatenate(mesh_elem_labs_all)
    
    # REMOVE duplicates
    # find unique node coordinates
    mesh_nodes_all_uniq, index, counts = np.unique(
        mesh_nodes_all, axis=0, 
        return_index=True, 
        return_counts=True
    )

    # labels of unique nodes
    mesh_node_labs_all_uniq = mesh_node_labs_all[index]

    # indices of removed nodes
    rem_nodes_index = np.setdiff1d(np.arange(len(mesh_nodes_all)), index)
    removed_nodes = mesh_nodes_all[rem_nodes_index]


    mesh_elem_all_uniq = mesh_elem_all.copy()
    if np.any(counts>2):
        raise Warning('Label repeated more than twice.')
    else:
        for ni, rep_node in enumerate(mesh_nodes_all_uniq[counts==2]):

            pl_idx_ni = np.nonzero((mesh_nodes_all_uniq==rep_node).all(axis=1))
            rep_nodes_idx = np.nonzero((removed_nodes==rep_node).all(axis=1))
            rep_nodes = removed_nodes[rep_nodes_idx][0]
            rep_nodes_lab = mesh_node_labs_all[rem_nodes_index][rep_nodes_idx][0]
            mesh_elem_all_uniq[np.where(mesh_elem_all==rep_nodes_lab)] = mesh_node_labs_all_uniq[pl_idx_ni]
    
    out = {
        'node_coordinates': mesh_nodes_all_uniq,
        'node_labels': mesh_node_labs_all_uniq,
        'element_nodes': mesh_elem_all_uniq,
        'element_labels': mesh_elem_labs_all,
    }
    
    return out


def compact_tension_specimen_mesh(refined_mesh_definition, dimensions, size_type, fraction):

    transition_mesh_definition = transition_mesh(refined_mesh_definition)
    crack_nodes = refined_mesh_definition['crack_definition']

    nd = transition_mesh_definition['node_coordinates']
    ndlab = transition_mesh_definition['node_labels']
    el = transition_mesh_definition['element_nodes']
    ellab = transition_mesh_definition['element_labels']
    
    specimen = fracture.standard_specimen(
        size_type,
        dimensions=dimensions,
        fraction=fraction
    )


    tot_size_ref = (
        np.max(nd[:,0]) - np.min(nd[:,0]),
        np.max(nd[:,1]) - np.min(nd[:,1]),
    )

    exp_width = (specimen['A'] - tot_size_ref[0]) / 2
    exp_height = (specimen['E'] - tot_size_ref[1])

    num_lays_width = int(exp_width // 1)
    width_lays_width = 1 + exp_width % 1 / num_lays_width

    num_lays_height = int(exp_height // 1)
    width_lays_height = 1 + exp_height % 1 / num_lays_height

    
    for i in range(num_lays_width):
        # find idx of border nodes
        conds = ['minx', 'maxx']

        # expand directions (normal to border)
        expand_dirs = [
            np.array([-1, 0]),
            np.array([1, 0])
        ]

        # start-end direction (parallel to border)
        end_dirs = [
            np.array([0, 1]),
            np.array([0, 1])
        ]

        nd, ndlab, el, ellab = expand_mult_dirs(conds, expand_dirs, end_dirs, 
                                                nd, ndlab, el, ellab,
                                                num_layers=[1, 1], width=width_lays_width)


    for i in range(num_lays_height):
        conds = ['maxy', ]

        # expand directions (normal to border)
        expand_dirs = [
            np.array([0, 1]),
        ]

        # start-end direction (parallel to border)
        end_dirs = [
            np.array([1, 0]),
        ]

        nd, ndlab, el, ellab = expand_mult_dirs(
            conds, expand_dirs, end_dirs,nd, ndlab, el, ellab,
            num_layers=[1,], width=width_lays_height
        )
    
    # ****** ADD LOADING HOLE *******
    circ_centre = (
         - (specimen['W'] - specimen['A'] / 2),
        specimen['F'] / 2
    )
    circ_rad = specimen['C'] / 2



    pts_on_circ = utils.circle_points(circ_rad, 80, circ_centre)
    pts_on_circ_ridge1 = utils.circle_points(circ_rad - 2, 80, circ_centre)

    # FIND NODES CLOSE TO HOLE
    atol = 13 
    nd_dist_circ = (nd - circ_centre)[:,0] ** 2 + (nd - circ_centre)[:,1] ** 2
    pts_in_idx = np.nonzero(nd_dist_circ < (circ_rad - 0.2) ** 2)
    pts_out_idx = np.nonzero(nd_dist_circ >= (circ_rad - 0.2) ** 2 )[0]
    pts_close_idx = np.nonzero(np.isclose(np.abs(nd_dist_circ - (circ_rad - 0.2) ** 2), 0, atol=atol))
    pts_out_close_idx = np.nonzero(
        np.logical_and(
            np.isclose(np.abs(nd_dist_circ - (circ_rad - 0.2) ** 2), 0, atol=atol),
    #         (pts_close_idx, nd_dist_circ >= (circ_rad - 0.2) ** 2 - 0.2)[1]
             nd_dist_circ >= (circ_rad - 0.2) ** 2 - 0.2
        )
    )[0]

    # nodes hole top
    ridge_el = []
    ridge_nd_top = []
    ridge_ndlab_top = []

    # loop through points near circle 
    for i, pt in enumerate(nd[pts_out_close_idx]):
        # distance between pt and pts on circle
        diff_vecs = pt - pts_on_circ
        dist = (diff_vecs[:,0] ** 2 + diff_vecs[:,1] ** 2) ** 0.5
        nd[pts_out_close_idx[i]] = pts_on_circ[np.argmin(dist)]

        if pts_on_circ[np.argmin(dist)][1] > 22.0:
            ridge_nd_top.append(pts_on_circ[np.argmin(dist)][None])
            ridge_ndlab_top.append(ndlab[pts_out_close_idx[i]])


    ridge_sort_idx = np.argsort(np.concatenate(ridge_nd_top, axis=0)[:,0])
    ridge_nd_top_sort = np.array(ridge_ndlab_top)[ridge_sort_idx]
    for i, n in enumerate(ridge_nd_top_sort[:-1]):
        ridge_el.append([
        np.max(ndlab)+1, np.max(ndlab)+1, ridge_nd_top_sort[i+1], n
    ])
    ridge_el = np.array(ridge_el)


    # REMOVE ELEMENTS
    el_h = np.copy(el)
    ellab_h = np.copy(ellab)

    for i, e in enumerate(el_h):
        if np.any(np.isin(e, ndlab[pts_in_idx])):
    #         print(np.nonzero(np.isin(e, ndlab[pts_in_idx]))[0])
    #         print(ellab_h[i], e)
            rem_idx = np.nonzero(np.isin(e, ndlab[pts_in_idx]))[0]
    #         print('rem_idx: ', rem_idx)
            if rem_idx.shape[0] == 1:
                if rem_idx[0] < len(e)-1:
                    e[rem_idx[0]] = e[rem_idx[0]+1]
                else:
                    e[rem_idx[0]] = e[0]
                el_h[i] = e

    rem_idx = np.concatenate((np.nonzero(np.count_nonzero(np.isin(el_h, ndlab[pts_in_idx]), axis=1)==2)[0],
    np.nonzero(np.count_nonzero(np.isin(el_h, ndlab[pts_in_idx]), axis=1)==3)[0],
    ))
    rem_idx = np.concatenate((rem_idx, np.nonzero(np.all((np.isin(el, ndlab[pts_in_idx])), axis=1))[0] ))


    el_h = np.delete(el_h, rem_idx, axis=0)
    ellab_h = np.delete(ellab_h, rem_idx, axis=0)

    # CREATE RIDGE
    nd = nd[pts_out_idx]
    ndlab = ndlab[pts_out_idx]
    nd = np.append(nd, np.array(circ_centre)[None], axis=0)
    # print(nd)
    ndlab = np.append(ndlab, np.max(ndlab)+1)
    ridge_centre_lab = ndlab[-1]


    specimen_set = (np.min(ellab_h), np.max(ellab_h))
    el_h = np.concatenate((el_h, ridge_el))

    ridge_el_set = np.arange(np.max(ellab_h)+1, np.max(ellab_h)+1+ridge_el.shape[0])
    ellab_h = np.concatenate((ellab_h, ridge_el_set))

    num_ridge_el = ridge_el.shape[0]
    num_nodes_layer = len(ndlab)
    loadline_nset = ndlab[-1]
    mid_plane_set = ndlab
    
    # **** Find BCs ****
    nodes_b, labs_b = find_border(nd, ndlab,
                                  condition='miny')
    
    out = {
        'node_coordinates': nd,
        'node_labels': ndlab,
        'element_nodes': el_h,
        'element_labels': ellab_h,
        'elsets':{
            'specimen': specimen_set, #defined as list or (start, stop, [step])
            'ridge': ridge_el_set,

        },
        'nsets': {
            'crackfront': labs_b,
            'load-line': loadline_nset,
    #         'cracktip': crack_nodes['crack_line'][:,0],
    #         'crackline': crack_nodes['crack_line'][:,1:].flatten(),
            'cracklip': crack_nodes['crack_lip'],
            'flank': np.concatenate((crack_nodes['crack_line'].flatten(), crack_nodes['crack_lip'])),
            'midplane': mid_plane_set,
        },
    }
    
    return out


def standard_specimen(spec_type, dimensions='2D', fraction='half', aw_ratio=0.5):
    """
    Returns the dimensions of standard fracture specimens.

    Parameters
    ----------
    spec_type
        CT-1T/SENB
    dimensions
        2D/3D
    fraction
        quarter/half/full
    aw_ratio
        fix at 0.5 for CT
    
    Returns
    -------
    if 2D:
        a/w, W, A, C, D, E, F
    if 3D:h
        a/w, W, A, B, C, D, E, F
    """
    
    specimens_dims = {
    'ct-1t':{
            'a/w': 0.5,
            'W': 50,
            'A': 62.5,
            'B': 25,
            'C': 12.5,
            'D': 23.5,
            'E': 60,
            'F': 37.5
        }
    }
    specimen = specimens_dims[spec_type]
    if dimensions=='2D':
        specimen.pop('B')
        
    if fraction=='half':
        specimen['E'] *= 0.5

    return specimens_dims['ct-1t']
    


def generate_compact_tension_specimen_steps(applied_displacement, number_contours, time_increment_definition):

    time_increment_def = (
        time_increment_definition['initial_time_increment'],
        time_increment_definition['total_step_time'],
        time_increment_definition['min_time_increment_allowed'],
        time_increment_definition['max_time_increment_allowed'],
    )
    steps = {
        'initial-step': {
            'bcs': [
                {
                'node set': 'load-line',
                'dof': (1, 1)
                },
                {
                'node set': 'load-line',
                'dof': (2, 2)
                },
                {
                    'node set': 'crackfront',
                    'type': 'YSYMM'
                },
                {
                    'node set': 'midplane',
                    'type': 'ZSYMM'
                }
            ]
        },
        'load-step-1': {
            'name': 'Step-1',
            'type': 'Static',
            'time_increment_definition': time_increment_def, #(0.02, 1.0, 1e-08, 0.02),
            'bcs':[{
                'node set': 'load-line',
                'dof': (2, 2, applied_displacement)
            },],
            'output':
            {
                'restart frequency': 0,
                'field':[
                    {
                        'output type': 'node',
                        'variables': ['COORD', 'U'],
                    },
                    {
                        'output type': 'element',
                        'position': 'centroidal',
                        'elset': 'specimen',
                        'frequency': 0,
                        'variables': ['E', 'EVOL', 'PE', 'PEEQ', 'S', 'COORD'],
                    }
                ],
                'history':
                {
                    'frequency': 0,
                    'cracks': [
                        {
                            'name': 'CRACK',
                            'contours': number_contours,
    #                         'crack tip nodes': [ 'crackline', 'cracktip'],
                            'crack tip nodes': [],
                            'symmetry': True,
                            'direction': [1, 0, 0],
                        }
                    ]
                },
            }
            
        }
    }

    return steps


import numpy as np

def get_euromat_A_material_props(temp):
    """Get mechanical properties of Euromaterial A, calculated as a function of
    temperature from a fit to uniaxial tests [1].
    
    Parameters
    ----------
    temp : float
        Temperature in Celsius.
    
    References
    ---------
    [1] Beardsmore, D. & Lidbury, D. (2006). Serco Assurance report. PERFECT.
    
    Returns
    -------
    material_props : dict
        Dict containing keys:
            yield_stress : float
                Yield stress in MPa.
            youngs_modulus : float
                Young's modulus in MPa.
            hardening_exponent : float
            UTS : float
                Ultimate tensile strength in MPa.
            poisson_ratio : float  
    """
    
    if temp >= 0:
        hard_exp = 9.96888 - 0.00185 * temp
    else:
        hard_exp = (9.96888 - 901.98738 * temp - 3.01593 * temp ** 2) / (1 - 90.48011 * temp - 0.41856 * temp ** 2)
    
    material_props = {
        'yield_stress': 421.18 + 63.9 * np.exp(-temp / 91),
        'youngs_modulus': -90 * temp + 206000,
        'hardening_exponent': hard_exp,
        'UTS': 564.1 + 70.2 * np.exp(-temp / 108),
        'poisson_ratio': 0.3,
    }
    
    return material_props


def get_euromat_A_material_definition(temp, plastic_stress_max=1200, plastic_num_incs=200):
    """
    Get an Abaqus compatible material definition for Euromaterial A.
    
    Paramaters
    ----------
    temp : float
        Temperature in Celsius.
    plastic_stress_max : float or int
        Maximum plastic stress in MPa.
    plastic_num_incs : int
        Number of increments of the plastic stress-strain return data.
    
    Returns
    -------
    material_defn : dict
        Dict containing keys:
            elastic : dict
                Dict containing keys:
                    youngs_modulus : float
                        Young's modulus in MPa.
                    poisson_ratio : float
            plastic : dict
                Dict containing keys:
                    stress_strain : ndarray of shape (N, 2)
                        Plastic stress-strain data; first column is stress.
    """
    
    euromat_props = get_euromat_A_material_props(temp)

    s0_wt = euromat_props['yield_stress']
    E_wt = euromat_props['youngs_modulus']
    n_wt = euromat_props['hardening_exponent']
    
    stress_data = np.linspace(s0_wt, plastic_stress_max, plastic_num_incs)
    tot_train = (s0_wt / E_wt) * (stress_data / s0_wt) ** n_wt
    pl_strain_data = tot_train - stress_data / E_wt
    
    material_defn =  {
        'Elastic': {
            'youngs_modulus' : E_wt,
            'poisson_ratio' : euromat_props['poisson_ratio'],
        }, 
        'Plastic': {
            'stress_strain': np.array([stress_data, pl_strain_data]).T
        }
    }
    return material_defn


import numpy as np

def k2j(k, E, nu, plane_stress=False):
    """
    Convert fracture 
    
    Parameters
    ----------
    k: float
    E: float
        Young's modulus in GPa.
    nu: float
        Poisson's ratio
    plane_stress: bool
        True for plane stress (default) or False for plane strain condition.
    
    Returns
    -------
    float
    
    """
    
    if plane_stress:
        E = E / (1 - nu ** 2)
        
    return k ** 2 / E


def j2k(j, E, nu, plane_stress=True):
    """
    Convert fracture 
    
    Parameters
    ----------
    j: float (in N/mm)
    E: float
        Young's modulus in GPa.
    nu: float
        Poisson's ratio
    plane_stress: bool
        True for plane stress (default) or False for plane strain condition.
    
    Returns
    -------
    K : float
        Units are MPa m^0.5.
    """
    
    if plane_stress:
        E = E / (1 - nu ** 2)
        
    return (j * E) ** 0.5

def calc_cpf(data):
    data_ordered = np.sort(data)
    u =  (np.arange(len(data)) + 1 - 0.3) / (len(data) + 0.4)
    
    return data_ordered, u

def find_closest(a, b, x, tol=0.01):
    """
    For two arrays `a` and `b`, find the elements in `b` closest to the `x` element in `a`.
    
    Parameters
    ----------
    a: narray of shape (N,)
    b: narray of shape (N,)
    x: float
    tol: float
        Tolerance value. 
    Returns
    -------
    """
    a_idx = np.isclose(a, x, atol=tol)
    a_i = a[a_idx]
    b_i = b[a_idx]

    return a_i, b_i
def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx

def Fk(shear_stress, temperature, Tath=700, Uk=0.33 * 1.60217657e-19, sigmap=900*1e6):
    """
    shear_stress : ndarray of shape (N,)
        CRSS or maximum shear stress (in Pa or Nm-2)
    temperature : float
        Temperature (in Kelvin)
    Tath : float
        The athermal temperature in Kelvin. Default set for iron = 700 K.
    Uk : float
        The activation energy (in Joules). Default set for iron.
    sigmap : float
    """

    fk = Uk *(1 - temperature / Tath - (shear_stress / sigmap) / (1 - temperature / Tath))
    fkneg_idx = np.where(fk<0)[0]
    fk[fkneg_idx] = 0.0
    
    return fk

def lstar(shear_stress, temperature, b=0.248e-9):
    """
    Parameters
    -------
    shear_stress : ndarray of shape (N,)
        CRSS or maximum shear stress (in Pa or Nm-2)
    temperature : float
        Temperature (in Kelvin)
    b : float
        Burger's vector (in m). Default set for iron. 
        
    Returns
    -------
    ndarray of shape (N,)
        Critical lenght L* (in m).
    
    """
    # Boltzmann constant
    kb =  1.3806498e-23 # J/K

    return b * np.exp( 1 / (kb * temperature) * Fk(shear_stress, temperature))


def thinning_function(obstacle_distance, shear_stress, temperature):
    """
    Thinning function applied in a modified local approach to cleavage fracture predictions.
    
    Parameters
    ----------
    obstacle_distance: array_like
        Average distance between dislocation obstacles [1].
    shear_stress: array_like
        Maximum shear stress in MPa
    temperature: float 
        Temperature in degree Celsius.
    Returns
    -------
    thinf : 

    References
    ----------
    [1] T. D. Swinburne and S. L. Dudarev, “Kink-limited Orowan strengthening explains the brittle to ductile transition of irradiated and unirradiated bcc metals,” Phys. Rev. Mater., vol. 2, p. 73608, 2018.
    
    """
    # Boltzmann constant
    kb =  1.3806498e-23 # Joules/Kelvin
    
    thinf = np.zeros(len(shear_stress))
    idx_more = lstar(shear_stress*1e6, temperature) > np.ones(len(shear_stress)) * obstacle_distance
    idx_less = ~idx_more
    
    # normalisation factor for free energy
    norm_factor =  1 / (
        2 * Fk(np.array([0]), -200+273.15) / (kb * (-200+273.15)))[0]

    thinf[idx_more] = norm_factor * 2 * Fk(shear_stress, temperature)[idx_more] / (kb * temperature) 
    thinf[idx_less] = norm_factor * Fk(shear_stress, temperature)[idx_less] / (kb * temperature)
        
    return thinf


from abaqus_parse.materials import euromat_A

MATERIAL_MODEL_FUNCS = {
    'euromat_A': euromat_A.get_euromat_A_material_definition,
}

def generate_material_models(materials_list):
    
    material_models = {}
    
    for material in materials_list:
        mat_mod_i = {}

        if 'func_name' in material:
            func = MATERIAL_MODEL_FUNCS[material['func_name']]
            func_kwargs = {key: value for key, value in material.items()
                           if key not in ['name', 'func_name']}
            mat_mod_i = func(**func_kwargs) 

        else:
            if 'elastic' in material:
                mat_mod_i['Elastic'] = material['elastic']

        material_models[material['name']] = mat_mod_i


    return material_models


import numpy as np
def flatten_list(a):
    return [item for sublist in a for item in sublist]

def format_args_check(**kwargs):
    """
    Check types of parameters used in `format_arr`, 'format_list' and
    'format_dict' functions.

    """

    if 'depth' in kwargs and not isinstance(kwargs['depth'], int):
        raise ValueError('`depth` must be an integer.')

    if 'indent' in kwargs and not isinstance(kwargs['indent'], str):
        raise ValueError('`indent` must be a string.')

    if 'col_delim' in kwargs and not isinstance(kwargs['col_delim'], str):
        raise ValueError('`col_delim` must be a string.')

    if 'row_delim' in kwargs and not isinstance(kwargs['row_delim'], str):
        raise ValueError('`row_delim` must be a string.')

    if 'dim_delim' in kwargs and not isinstance(kwargs['dim_delim'], str):
        raise ValueError('`dim_delim` must be a string.')

    if 'format_spec' in kwargs and not isinstance(kwargs['format_spec'],
                                                  (str, list)):
        raise ValueError('`format_spec` must be a string or list of strings.')

    if 'assign' in kwargs:

        if not isinstance(kwargs['assign'], str):
            raise ValueError('`assign` must be a string.')

def format_arr(arr, depth=0, indent='\t', col_delim='\t', row_delim='\n',
               dim_delim='\n', format_spec='{}'):
    """
    Get a string representation of a Numpy array, formatted with indents.

    Parameters
    ----------
    arr : ndarray or list of ndarray
        Array of any shape to format as a string, or list of arrays whose
        shapes match except for the final dimension, in which case the arrays
        will be formatted horizontally next to each other.
    depth : int, optional
        The indent depth at which to begin the formatting.
    indent : str, optional
        The string used as the indent. The string which indents each line of
        the array is equal to (`indent` * `depth`).
    col_delim : str, optional
        String to delimit columns (the innermost dimension of the array).
        Default is tab character, \t.
    row_delim : str, optional
        String to delimit rows (the second-innermost dimension of the array).
        Default is newline character, \n.
    dim_delim : str, optional
        String to delimit outer dimensions. Default is newline character, \n.
    format_spec : str or list of str, optional
        Format specifier for the array or a list of format specifiers, one for 
        each array listed in `arr`.

    Returns
    -------
    str

    """

    # Validation:
    format_args_check(depth=depth, indent=indent, col_delim=col_delim,
                      row_delim=row_delim, dim_delim=dim_delim,
                      format_spec=format_spec)

    if isinstance(arr, np.ndarray):
        arr = [arr]

    out_shape = list(set([i.shape[:-1] for i in arr]))

    if len(out_shape) > 1:
        raise ValueError('Array shapes must be identical apart from the '
                         'innermost dimension.')

    if not isinstance(arr, (list, np.ndarray)):
        raise ValueError('Cannot format as array, object is '
                         'not an array or list of arrays: type is {}'.format(
                             type(arr)))

    if isinstance(format_spec, str):
        format_spec = [format_spec] * len(arr)

    elif isinstance(format_spec, list):

        fs_err_msg = ('`format_spec` must be a string or list of N strings '
                      'where N is the number of arrays specified in `arr`.')

        if not all([isinstance(i, str)
                    for i in format_spec]) or len(format_spec) != len(arr):
            raise ValueError(fs_err_msg)

    arr_list = arr
    out = ''
    dim_seps = ''
    d = arr_list[0].ndim

    if d == 1:
        out += (indent * depth)

        for sa_idx, sub_arr in enumerate(arr_list):
            for idx_col, col in enumerate(sub_arr):
                if idx_col==0 and sa_idx==0 and idx_col==len(sub_arr)-1 and sa_idx==len(arr_list)-1:
                    out += '   ' + format_spec[sa_idx].format(col)
                elif idx_col==len(sub_arr)-1 and sa_idx==len(arr_list)-1:
                    out += format_spec[sa_idx].format(col)
                elif idx_col==0 and sa_idx==0:
                    out += '   ' + format_spec[sa_idx].format(col) + col_delim
                else:
                    out += format_spec[sa_idx].format(col) + col_delim
     
        out += row_delim

    else:

        if d > 2:
            dim_seps = dim_delim * (d - 2)

        sub_arr = []
        for i in range(out_shape[0][0]):

            sub_arr_lst = []
            for j in arr_list:
                sub_arr_lst.append(j[i])

            sub_arr.append(format_arr(sub_arr_lst, depth, indent, col_delim,
                                      row_delim, dim_delim, format_spec))

        out = dim_seps.join(sub_arr)

    return out

def circle_points(r, n, centre):
    t = np.linspace(0, 2*np.pi, n)
    x = r * np.cos(t) + centre[0]
    y = r * np.sin(t) + centre[1]

    return np.c_[x, y]

def polar2cart_2D(r, θ):
    """
    Convert 2D polar coordinates to Cartesian coordinates.

    """

    x = r * np.cos(θ)
    y = r * np.sin(θ)

    return x, y


def order_coplanar_points(points, normal, anticlockwise=True):
    """
    Find the clockwise or anticlockwise ordering of a set of coplanar 3D points.

    Parameters
    ----------
    points : ndarray of shape (3, N)
        The set of coplanar points (three-vector columns) whose ordering is to be found.
    normal : ndarray of shape (3, 1)
        Column three-vector representing the normal of the plane on which all points lie.

    Returns
    -------
    Ordered indices of points according to a clockwise or anticlockwise direction when
    looking in the opposite direction to `normal`.

    """

    # Normalise `normal` to a unit vector:
    normal = normal / np.linalg.norm(normal)

    # Compute the centroid:
    centroid = np.mean(points, axis=1)

    # Get the direction vectors from each point to the centroid
    p = points - centroid[:, np.newaxis]

    # Use the first point as the reference point
    # Find cross product of each point with the reference point
    crs = np.cross(p[:, 0:1], p, axis=0)

    # Find the scalar triple product of the point pairs and the normal vector
    stp = np.einsum('ij,ik->k', normal, crs)

    # Find the dot product of the point pairs
    dot = np.einsum('ij,ik->k', p[:, 0:1], p)

    # Find signed angles from reference point to each other point
    ang = np.arctan2(stp, dot)
    ang_order = np.argsort(ang)

    if not anticlockwise:
        ang_order = ang_order[::-1]

    return ang_order



import numpy as np
from abaqus_parse.utils import format_arr
from abaqus_parse._version import __version__
__all__ = []

def write_inp(path, materials, parts, steps, assembly=None):
    """
    Write the input file '.inp' for an Abaqus simulation.
    
    Parameters
    ----------
    path : string
        Path for writing the input file.
    materials : dict of dicts
         Dict whose keys are the material names and whose values are dicts with the following keys:
            elastic : dict, optional
                Dict with keys:
                    youngs_modulus : float
                    poisson_ratio : float
            func_name : dict, optional, a function that generates a dictionary with the material properties, e.g. 
                elastic :  dict
                    Dict with keys:
                        youngs_modulus : float
                        poisson_ratio : float
                plastic : dict
                    stress_strain : ndarray of shape (R, 2)
            **func_kwargs, optional, input arguments for func_name, e.g. temperature.
    parts : dict of dicts
        Dict whose keys are the part names and whose values are dicts with the following keys:
            node_coordinates : ndarray of shape (M, N) for M nodes with N coordinates
            node_labels : ndarray of shape (M,) for M node labels
            element_nodes : ndarray of shape (L, K) for L elements with K nodes each
            element_labels : ndarray of shape (L,) for L label nodes
            nsets : dict
                Dict keys are the node set names and values are one of:
                    - range of nodes specified as a tuple (start, stop, [step]), OR
                    - ndarray of shape (P,) for P nodes, OR
                    - integer for a single node
            elsets : dict
                Dict keys are the element set names and values are one of:
                    - range of elements specified as a tuple (start, stop, [step]), OR
                    - ndarray of shape (Q,) for Q elements, OR
                    - integer for a single element
            sections : list of dicts
                Each dictionary has the following keys and values:
                    type : str
                        The type of section, e.g. 'Solid'.
                    material : str
                        The name of the material of the section.
                    elset : str
                        The name of the element set part of the section. 
    assembly : list of dicts
        Definitions of the part instances that the assembly consists of (optional) (TODO).
    steps: dict of dicts 
        Dict whose keys are the step names and whose values are dicts with the following keys. The first step name must be 'initial-step'.
            initial-step : dict 
                Dict with keys:
                    bcs : list of dict
                        List of boundary condition dicts, each with keys:
                            node set : str 
                                The node set name
                            dof : tuple, optional 
                                Degrees of freedom to be included into the boundary condition
                            type : str, optional
                                Type of boundary condition, e.g. 'XSYMM' OR 'PINNED'.
            analysis_step_name : dict
                Dict representing arbitrary step with keys:
                    name : str
                    type : str
                    time_increment_definition : tuple 
                        Representing (initial_time_increment, total_step_time, min_time_increment_allowed, max_time_increment_allowed)
                    bcs : list of dict
                        List of boundary condition dicts, each with keys:
                            node set : str 
                                The node set name
                            dof : tuple, optional 
                                Degrees of freedom to be included into the boundary condition
                    output : dict of dicts 
                        history : dict
                            frequency : int, optional
                            cracks : list of dicts, optional
                        field : list of dicts
                            output type : str, node | element
                            position : str, optional, e.g. 'centroidal'
                            set name : str, optional
                            variables : list of str, e.g. ['COORD', 'U']
                      
    Returns
    -------
    An Abaqus .inp file.

    TODO:
    - Add preprint options as input parameters. Currently, hard coded.
    - User specified Field Output. Currently, hard coded.
    - Change steps to to list of dicts

    """
    # ********** PARTS **********

    for part, part_definition in parts.items():

        node_coordinates = part_definition['node_coordinates']
        node_labels = part_definition['node_labels']
        elem_nodes = part_definition['element_nodes']
        elem_labs = part_definition['element_labels']
        elem_type = part_definition['element_type']
        elem_sets = part_definition['elsets']
        node_sets = part_definition['nsets']
        sections = part_definition['sections']

        sep = '\n'

        nodes = [
            '**\n',
            '**Nodes\n',
            '**\n',
            '*Node\n',
            format_arr([node_labels[None].T, node_coordinates],
                                format_spec=['{:d}', '{:12.7f}'],
                                col_delim=', ')  
        ]
        
        elems = [
            '**\n',
            '**Elements\n',
            '**\n',
            '*Element, type=' + elem_type + '\n',
            format_arr([elem_labs[None].T, elem_nodes],
                                format_spec=['{:d}', '{:d}'],
                                col_delim=', ')
        ]
        
        n_sets = [
            '**\n',
            '**Node sets\n',
            '**\n',
        ]
        for k, v in node_sets.items():
            if type(v) == tuple:
                n_sets.append(
                    '*Nset, nset=' + k + ', generate\n' +\
                    str(v[0]) + ', ' + str(v[1]) + ', 1\n' 
                            )
            elif type(v)==list or type(v)==np.ndarray:
                if type(v)==list:
                    v = np.array(v)
                whole_rows = v.size // 16
                first_block = v[:(whole_rows * 16)].reshape(-1, 16)
                remaining_block = v[(whole_rows * 16):]
                n_sets.append('*Nset, nset=' + k + '\n' +\
                            format_arr(first_block, format_spec=['{:d}'],
                                    col_delim=', ') +\
                            format_arr(remaining_block, format_spec=['{:d}'],
                                    col_delim=', '))
            elif type(v)==np.int32 or type(v)==np.int64:
                n_sets.append('*Nset, nset=' + k + '\n' + str(v) + '\n') 
        el_sets = [
            '**\n',
            '**Element sets\n',
            '**\n',
        ]
        for k, v in elem_sets.items():
            if type(v) == tuple:
                el_sets.append(
                    '*Elset, elset=' + k + ', generate\n' +\
                    str(v[0]) + ', ' + str(v[1]) + ', 1\n' 
                            )
            elif type(v)==list or type(v)==np.ndarray:
                whole_rows = v.size // 16
                first_block = v[:(whole_rows * 16)].reshape(-1, 16)
                remaining_block = v[(whole_rows * 16):]
                n_sets.append('*Elset, elset=' + k + '\n' +\
                            format_arr(first_block, format_spec=['{:d}'],
                                    col_delim=', ') +\
                            format_arr(remaining_block, format_spec=['{:d}'],
                                    col_delim=', '))
            elif type(v)==np.int32 or type(v)==np.int64:
                n_sets.append('*Elset, elset=' + k + '\n' + str(v) + '\n') 
        
        # Sections
        sects = [
            '**\n',
            '**Sections\n',
            '**\n',
        ]
        for sect in sections:   
            sects.append(
                '*' + sect['type'] + ' Section, elset=' + sect['elset']  +\
                ', material=' + sect['material'] + '\n'
            )
        sects = sep.join([
            ''.join(sects),
        ])
     # ********** MATERIALS **********
    mats = [
        '**\n',
        '**Materials\n',
        '**\n',
    ]
    for k, v in materials.items():
        mats.append(
            '*Material, name=' + k + '\n')
        for sk, ss in v.items():
            if sk == 'Elastic':
                ss = [[ss['youngs_modulus'], ss['poisson_ratio']]]
            if sk == 'Plastic':
                ss = ss['stress_strain']
            mats.append(
                '*' + sk + '\n' +\
                format_arr(np.array(ss), format_spec=['{:12.7f}'],
                            col_delim=', ')
            )
            
    # ********** STEPS **********       
    stps = [
        '**\n',
        '**Boundary conditions\n',
        '**\n',
    ]
    for k, v in steps.items():
        if k != 'initial-step':
            
            stps.append(
                '*Step, name=' + v['name'] + ', nlgeom=YES\n' +\
                '*' + v['type'] + '\n' + format_arr(list(np.array(list(v['time_increment_definition']))[None].T), format_spec=['{:3.2f}', '{:3.1f}','{:2.1e}', '{:3.2f}'],
                                                   col_delim=', ')
            )
        for bc in v['bcs']:
                stps.append(
                    '*Boundary\n' + bc['node set'] +', ')
                if 'dof' in bc.keys():
                    if len(bc['dof'])==2:
                        stps.append(format_arr(np.array(bc['dof']), format_spec=['{:d}'], col_delim=', ') )
                    else:
                        stps.append(str(bc['dof'][0]) + ', ' + str(bc['dof'][1])+ ', ' + str(bc['dof'][2]) + '\n')
                elif 'type' in bc.keys():
                    stps.append(bc['type'] + '\n')
        if 'output' in v.keys():
            if 'restart frequency' in v['output'].keys():
                stps.append(
                    '*Restart, write, frequency=' + str(v['output']['restart frequency']) +'\n'
                )
            for ko, vo in v['output'].items():
                if ko=='field':
                    stps.append(
                        '*Output, field\n'
                    )
                    for fo in vo:
                        if fo['output type']=='node':
                            stps.append('*Node Output')
                            if 'set name' in fo.keys():
                                stps.append(', nset='+fo['set'])
                            if 'frequency' in fo.keys():
                                stps.append(', frequency='+str(fo['frequency']))
                            stps.append('\n' + ', '.join(fo['variables']) + '\n')
                        if fo['output type']=='element':
                            stps.append('*Element Output')
                            if 'position' in fo.keys():
                                stps.append(', position='+fo['position'])
                            if 'set name' in fo.keys():
                                stps.append(', elset='+fo['set'])
                            if 'frequency' in fo.keys():
                                stps.append(', frequency='+str(fo['frequency']))
                            stps.append('\n'+ ', '.join(fo['variables']) + '\n')     
                elif ko=='history':
                    stps.append(
                        '*Output, history, frequency=' + str(vo['frequency']) + '\n'
                    )
                    # if 'cracks' in vo.keys():
                    #     for crack in vo['cracks']:
                    #         stps.append(
                    #             '*Contour Integral, crack name=' + str(crack['name']) + ', contours=' + str(crack['contours'])
                    #         )
                    #         if 'crack tip nodes' in crack.keys():
                    #             stps.append(', crack tip nodes')
                    #             if crack['symmetry']:
                    #                 stps.append(', symm')
                    #             stps.append('\n')
                    #             if any(isinstance(el, list) for el in crack['crack tip nodes']):
                    #                 for cr in crack['crack tip nodes']:
                    #                      stps.append(cr[0] + ', ' + cr[1] + ', ' + format_arr(np.array(crack['direction']), format_spec=['{:d}'], col_delim=', '))
                    #             else:
                    #                 stps.append('\n' + crack['crack tip nodes'][0] + ', ' + crack['crack tip nodes'][1] + ', ' + format_arr(np.array(crack['direction']), format_spec=['{:d}'], col_delim=', '))
              
                            
        if k != 'initial-step':       
            stps.append('*End Step\n')
    
    with open(path, 'w') as of:   # to do: test if i need universal newline mode here

        # Input file heading
        of.write('*Heading\n')
        of.write('** Generated by: abaqus-parse v' + str(__version__) +' \n')
        of.write('*Preprint, echo=NO, model=NO, history=NO, contact=NO\n')
        of.write(
            ''.join([
                ''.join(nodes),
                ''.join(elems),
                ''.join(n_sets),
                ''.join(el_sets),
                ''.join(sects),
                ''.join(mats),
                ''.join(stps),
            ])
        )

   


# code

data_labels = ['label', 'xcoord', 'ycoord', 'zcoord', 'vol', 's1', 's2', 's3', 'p1', 'p2', 'p3', 'peeq', 'sh', 'seq']
front_data_all = []
skipped_lines_all = []
for ip, path in enumerate(paths):
    if calc_bool[ip]:
        temp_ip = temps[ip] + 273
        front_data_all.append({
            'file_path': file_path,
            'data': {},
        })
        nodes_ip = front_coords_all[ip][0]
        elements_ip = front_coords_all[ip][1]
        nsets_ip = front_coords_all[ip][2]

        nodes_front = {
            'index': [],
        }
        elements_front = {
            'index': [],
            'nodal connectivity': []
        }
        good_idx = np.where(np.isin(nodes_ip['index'], nsets_ip[front_name]))[0]
        nodes_front['coord'] = nodes_ip['coord'][good_idx].squeeze()
        nodes_front['index'] = nsets_ip[front_name]

#         print('nodes_front', nodes_front)
        # # find elements with 2 nodes that belong to above set 
        for row_idx, i in enumerate(elements_ip['nodal connectivity']):
            isin = np.isin(i, nodes_front['index'])
#             print(isin)
            if row_idx == 2:
                print(np.sum(isin))
            if np.sum(isin) == 4:
        #         print('good row idx: {}'.format(row_idx))
                elements_front['index'].append(elements_ip['index'][row_idx])
                elements_front['nodal connectivity'].append(i)
                
        num_front_elem = len(np.array(elements_front['index'])) // num_layers[ip]

        elements_front['index'] = np.array(elements_front['index'])[0:num_front_elem]
        elements_front['nodal connectivity'] = np.array(elements_front['nodal connectivity'])[0:num_front_elem]

        curr_line = 0
        skipped_lines = 0
        front_data_all[ip]['data'] = []
        curr_line += 1
        labels_0 = pd.Series(dtype=np.float)
        data_0 = 0
            
        for file_path in glob.iglob(path+'\ehis*'):
            print(file_path)
            with open(file_path, 'rb') as file:
                line = file.readline()

            frame_i, num_elem_i = [int(n) for n in line.strip().split()]
            if num_elem_i == 0:
                skipped_lines += 1
            elif num_elem_i > 0:
                data_i = pd.read_csv(file_path, delim_whitespace=True, skiprows=1, index_col=False,
                                                 nrows=num_elem_i, names=data_labels)
                labels_i = data_i['label']
                front_labels_i_idx = np.where(np.isin(labels_i, elements_front['index']))[0]

                if len(front_labels_i_idx)>0:
    #                     print('front_labels_i_idx: ', front_labels_i_idx)
                    pl_elem_check = labels_i.isin(labels_0)
                    x_i = data_i['xcoord'].values
                    y_i = data_i['ycoord'].values
                    smin_i = np.min(data_i[['s1', 's2', 's3']], axis=1)
                    smax_i = np.max(data_i[['s1', 's2', 's3']], axis=1)
                    pmax_i = np.max(data_i[['p1', 'p2', 'p3']], axis=1)
                    if 'blm' in path:
                        vols_i = data_i['vol'] * 1e9
                    else:
                        vols_i = data_i['vol']

                    peeq_i = data_i['peeq']
                    sh_i = data_i['sh']
                    seq_i = data_i['seq']
                    tau_i = (smax_i - smin_i) / 2
                    thinf_i = thinf(dbrack, tau_i, temp_ip)


    #                     smax_i_allt = smax_i.copy()
    #                     smax_i_allt[labels_i.isin(labels_0)] = np.max([smax_0.values[np.where(labels_0.isin(labels_i))], 
    #                                                               smax_i_allt[labels_i.isin(labels_0)].values], axis=0)

                    peeq_0 = np.zeros(len(labels_i))

                    front_data_all[ip]['data'].append({
                        'frame_i': frame_i,
                        'num_elem_i': num_elem_i,
                        'elem_labels': labels_i[front_labels_i_idx],
                        'xcoord': x_i[front_labels_i_idx] - x_i[front_labels_i_idx][0],
                        'ycoord': y_i[front_labels_i_idx] - y_i[front_labels_i_idx][0],
                        'smax': smax_i[front_labels_i_idx],
                        'smin': smin_i[front_labels_i_idx],
    #                         'smaxallt': smax_i_allt[front_labels_i_idx],
                        'vol': vols_i[front_labels_i_idx],
                        'peeq': peeq_i[front_labels_i_idx],
                        'thinf': thinf_i[front_labels_i_idx],
                    })


                labels_0 = labels_i.copy()
    #                     smax_0 = smax_i_allt.copy()
                data_0 = data_i
        skipped_lines_all.append(skipped_lines)
